from dataclasses import Field,dataclass

@dataclass
class FieldInput:
    field_val : str
    field_name : str = None
    
def collect_input(field : Field,is_valid_fxn=None):
    is_valid_input = False
    new_val=''
    while not is_valid_input:
        new_val=input(f'{field.name}: ')
        if is_valid_fxn is None:
            is_valid_input=True
        else:
            is_valid_input=is_valid_fxn(FieldInput(field_name=field.name,field_val=new_val))
    return new_val
    
def is_valid_non_empty(field_input : FieldInput):
    field_val = field_input.field_val
    if field_val is None or field_val=='':
        return False
    return True


import json
from jbg_util.basic_util import get_instance_from_dict
from jbg_util.input_util import collect_input
import os 
from pathlib import Path
import shutil
from dataclasses import fields


class FileManager:
    @staticmethod        
    def save_obj_to_file(file_path : str, d):
        with open(file_path,'wt',encoding='utf-8') as f:
            f.write(json.dumps(d))
    
    @staticmethod
    def get_obj_from_file(file_path : str):
        with open(file_path,'rt',encoding='utf-8') as f:
            lines=f.read()
            if lines=='':
                return {}
            return json.loads(lines)
            
    @staticmethod
    def get_instance_from_file(file_path : str, cls):
        d = FileManager.get_obj_from_file(file_path)
        return get_instance_from_dict(d,cls)
        
    @staticmethod
    def get_instances_from_file(file_path : str, cls):
        dicts = FileManager.get_obj_from_file(file_path)
        return [get_instance_from_dict(d,cls) for d in dicts]
       
    @staticmethod
    def save_str_to_file(file_path : str, s):
        with open(file_path,'wt',encoding='utf-8') as f:
            return f.write(s)

    @staticmethod
    def get_str_from_file(file_path : str):
        with open(file_path,'rt',encoding='utf-8') as f:
            return f.read()    
            
    @staticmethod
    def delete_all_files_in_folder(folder_path):
        if os.path.exists(folder_path):
            shutil.rmtree(folder_path)
        FileManager.ensure_folder_exists(folder_path)
        
    @staticmethod
    def ensure_folder_exists(folder_path):
        Path(folder_path).mkdir(parents=True, exist_ok=True)
        
    @staticmethod
    def ensure_file_exists(file_path):
        if not os.path.exists(file_path):
            FileManager.save_str_to_file(file_path,'')
        
    @staticmethod
    def collect_and_save_input_for_undefined_fields(file_path : str, cls, is_valid_fxn=None):
        map = FileManager.get_obj_from_file(file_path)
        for field in fields(cls):
            if field.name not in map:
                new_val=collect_input(field, is_valid_fxn)
                map[field.name]=new_val
        FileManager.save_obj_to_file(file_path,map)
        return get_instance_from_dict(map,cls) 
        
    @staticmethod
    def collect_and_save_input(file_path : str, cls = None):
        map = FileManager.get_obj_from_file(file_path)
        new_map={}
        for key,dflt_val in map.items():
            new_val=input(f'{key} (default={dflt_val}): ')
            if new_val=='':
                new_val=dflt_val
            new_map[key]=new_val
        FileManager.save_obj_to_file(file_path,new_map)  
        
        if cls is not None:
            return get_instance_from_dict(new_map,cls)
        return new_map
        
    @staticmethod
    def get_row_count_in_file(file_path : str) -> int:
        with open(file_path,'rt',encoding='utf-8') as f:
            return len(f.readlines())


from pathlib import Path
class AnsiEscapes:
    RED = '\033[91m'
    GREEN = '\033[92m'
    YELLOW = '\033[93m'
    BLUE = '\033[94m'
    PURPLE = '\033[95m'
    CYAN = '\033[96m'
    WHITE = '\033[97m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'
    ENDC = '\033[0m'

class Logger:
    SECTION_DIVIDER_WIDTH=50
    
    def __init__(self,file_path=None,should_print_to_screen=False):
        self.file_path=None
        if file_path is not None:
            self.file_path = file_path
            Path(file_path).parent.mkdir(parents=True, exist_ok=True) #ensure the directory exists
                
        self.should_print_to_screen = should_print_to_screen
        
    def empty_log(self):
        if self.file_path is None:
            return
        with open(self.file_path,'wt',encoding='utf-8') as f:
            f.write('')

    def log_colored(self,color,*strs):
        colored_strs=[]
        for s in strs:
            s = f"{color}{s}{AnsiEscapes.ENDC}"
            colored_strs.append(s)
        
        if self.should_print_to_screen:
            print(*colored_strs)
            
        if self.file_path is not None:
            with open(self.file_path,'at',encoding='utf-8') as f:
                f.write(' '.join(strs)+'\n')

    def success(self,*s):
        self.log_colored(AnsiEscapes.GREEN,*s)

    def error(self,*s):
        self.log_colored(AnsiEscapes.RED,*s)

    def warning(self,*s):
        self.log_colored(AnsiEscapes.YELLOW,*s)

    def info(self,*s):
        self.log_colored(AnsiEscapes.CYAN,*s)

    def primary(self,*s):
        self.log_colored(AnsiEscapes.BLUE,*s)

    def secondary(self,*s):
        self.log_colored(AnsiEscapes.PURPLE,*s)
        
    def bold(self,*s):
        self.log_colored(AnsiEscapes.BOLD,*s)
        
    def underline(self,*s):
        self.log_colored(AnsiEscapes.UNDERLINE,*s)
        
    def log(self,*s):
        self.log_colored(AnsiEscapes.WHITE,*s)
        
    def section_header(self,title,text_color : AnsiEscapes = AnsiEscapes.WHITE, divider_color : AnsiEscapes = AnsiEscapes.WHITE):
        divider = ''.join(['*' for _ in range(0,Logger.SECTION_DIVIDER_WIDTH)])
        double_divider = divider+'\n'+divider
        self.log_colored(divider_color,'\n'+double_divider)
        self.log_colored(text_color,title)
        self.log_colored(divider_color,double_divider)
        
    def subsection_header(self,title,text_color : AnsiEscapes = AnsiEscapes.WHITE, divider_color : AnsiEscapes = AnsiEscapes.WHITE):
        divider = ''.join(['-' for _ in range(0,Logger.SECTION_DIVIDER_WIDTH)])
        self.log_colored(divider_color,'\n'+divider)
        self.log_colored(text_color,title)
        self.log_colored(divider_color,divider)


from datetime import datetime,timedelta
from tabulate import tabulate
from traceback import print_exc
from dataclasses import fields
from distutils.util import strtobool
TYPES = {cls.__name__: cls for cls in [str, int, float, bool, dict]}
TYPES['bool']=strtobool #allows us to convert the string input to a bool in a more intuitive way

def add_to_dte(dte,days=0, seconds=0, microseconds=0, milliseconds=0, minutes=0, hours=0, weeks=0):
    return dte + timedelta(days=days, seconds=seconds, microseconds=microseconds, milliseconds=milliseconds, minutes=minutes, hours=hours, weeks=weeks)

def dte_to_str(dte):
    if dte is None:
        return None
    return dte.strftime("%Y-%m-%d")
    
def now():
    return datetime.utcnow()

def first(lst,matching_fxn):
    try:
        return next(item for item in lst if matching_fxn(item))
    except StopIteration as e:
        return None
        
# A decorator that catches exceptions and logs them. 
# Requires that the function be an instance method and the instance has a self.logger.error() function
def catches_exceptions(fxn):
    def wrapper(self,*args,**kwargs):
        try:
            fxn(self,*args,**kwargs)
        except Exception as e:
            self.logger.error(f'Exception occurred during {fxn.__qualname__}: {str(e)}') 
            print_exc()
    return wrapper

# Accepts a dict and returns the dict data as an object of type cls. 
# Note, cls must be a dataclass or other class insantiatiable with keyword arguments.
def get_instance_from_dict(d : dict, cls):
    converted_dict = {}
    for field in fields(cls):
        val, type_name=None, None
        if field.name in d:
            val = d[field.name]
            
            type_name=field.type
            if type(type_name)!=str: #for some odd reason, type_name is sometimes a string literal and sometimes a class object (e.g. sometimes the string 'int' and sometimes the class int)
                type_name = type_name.__name__
        
        if val is not None:
            converted_dict[field.name]=TYPES[type_name](val)
        else:
            converted_dict[field.name]=None
    return cls(**converted_dict)
    
def print_table(rows,col_headers):
    rows = [col_headers, *rows]
    print(tabulate(rows, headers='firstrow', tablefmt='fancy_grid'))
    
def get_mapped_val(map,key):
    if key not in map:
        raise ValueError(f'unrecognized key: {key}')
    return map[key]